import pandas as pd
import matplotlib.pyplot as plt
import numpy as np

# Colors: https://matplotlib.org/examples/color/named_colors.html
# Types: https://vignette3.wikia.nocookie.net/robloxpokemonbrickbronze/images/5/5f/Type.png/revision/latest?cb=20170426085006
types = {"Bug":"olive","Electric":"yellow","Fire":"red","Grass":"lime","Normal":"slategray","Rock":"darkgoldenrod","Dark":"black","Fairy":"lightpink","Flying":"skyblue","Ground":"saddlebrown","Poison":"darkorchid","Steel":"lightsteelblue","Dragon":"mediumpurple","Fighting":"brown","Ghost":"rosybrown","Ice":"cyan","Psychic":"magenta","Water":"blue"}

"""vulnerableTypes = {"Bug":["Flying","Rock"],"Electric":["Ground"],"Fire":["Ground","Rock","Water"],"Grass":["Flying","Poison","Bug","Fire","Ice"],"Normal":["Fighting","Ghost"],"Rock":["Fighting","Ground","Steel","Water","Ice"],"Dark":["Fighting","Bug","Fairy"],"Fairy":["Poison","Steel"],"Flying":["Rock","Electric","Ice"],"Ground":["Water","Grass","Ice"],"Poison":["Ground","Psychic"],"Steel":["Fighting","Ground","Fire"],"Dragon":["Ice","Dragon","Fairy"],"Fighting":["Flying","Psychic","Fairy"],"Ghost":["Ghost","Dark"],"Ice":["Fighting","Rock","Steel","Fire"],"Psychic":["Bug","Ghost","Dark"],"Water":["Grass","Electric"]}"""

effectiveTypes = {"Bug":["Grass","Psychic","Dark"],"Electric":["Flying","Water"],"Fire":["Bug","Steel","Grass","Ice"],"Grass":["Ground","Rock","Water"],"Normal":[],"Rock":["Flying","Bug","Fire"],"Dark":["Ghost","Psychic"],"Fairy":["Fighting","Dragon","Dark"],"Flying":["Bug","Fighting","Grass"],"Ground":["Poison","Rock","Steel","Fire","Electric"],"Poison":["Grass","Fairy"],"Steel":["Rock","Ice","Fairy"],"Dragon":["Dragon"],"Fighting":["Normal","Rock","Steel","Ice","Dark"],"Ghost":["Ghost","Psycho"],"Ice":["Flying","Ground","Grass","Dragon"],"Psychic":["Fighting","Poison"],"Water":["Ground","Rock","Fire"]}


def make_autopct(data):
    def my_autopct(pct):
        val = int(round(pct*sum(data)/100.0))
        return '{v:d}'.format(v=val)
    return my_autopct

def calculateTotalCombatPower(data):
	""" Calculate the total combat power for each pokemon and add it to the dataset. 
	The total CP calculates as the sum of all combat stats (hp, attack, defense, etc.)
	"""
	data['Total']= data['HP'] + data['Attack'] + data['Defense'] + data['Sp. Atk'] + data['Sp. Def'] + data['Speed']
	print(data)
	return data
	
def printCSV(data, filename):
	""" Write the data to the specified file as csv.
	"""
	data.to_csv(filename,index=False,header=True)
	
def topNPokemon(data, n):
    """ Find the strongest n Pokemon and print them in a bar graph. Give each bar a color fitting the type of the pokemon"""
    pokemon = data.sort_values(['Total'])[-n:] # sort the pokemon by their total strength and take the last n rows (the n strongest pokemon)
    colors = []
    for type in pokemon['Type 1']:
        print(type)
        colors.append(types[type]) # filling the list with colors which will be used to color the bars in the bar graph. each pokemon type has its own color
        
    ax = pokemon.plot(x="Name", y="Total", kind="bar", color=colors, legend = False) # plotting the names on the x-axis, the total strength on the y-axis, and using the corresponding color as the bar color
    plt.title("Top "+str(n)+" strongest Pokémon")
    ax.set_xlabel("Pokémon")
    ax.set_ylabel("Total combat power")
    
def topNNoMegaPokemon(data, n):
    """ Find the strongest n non-mega Pokemon and print them in a bar graph. Removing all Mega-Forms can be achieved by removing all Pokemon with
	duplicate numbers."""
    data=data.drop_duplicates(subset='#') # dropping duplicates in the column named '#' to get rid of the 'Mega Evolutions' (they have the same number, but different strength)
    pokemon = data.sort_values(['Total'])[-n:]
    colors = []
    for type in pokemon['Type 1']:
        colors.append(types[type])
    ax = pokemon.plot(x="Name", y="Total", kind="bar", color=colors, legend = False)
    plt.title("Top "+str(n)+" strongest Pokémon (no Mega Evolutions)")
    ax.set_xlabel("Pokémon")
    ax.set_ylabel("Total combat power")
	
def averagePowerPerGeneration(data):
	""" Calculate the average powerlevel of Pokemon from each Generation, and print it to the commandline
	"""
	data=data.drop_duplicates(subset='#') # dropping duplicates in the column named '#'
	average = data.groupby("Generation").aggregate(np.mean).reset_index() # grouping the pokemon by their generation, calculating the mean and resetting the index
	print(average[["Generation","Total"]].rename(columns = {'Total':'Average combat power'})) # renaming column 'Total' to 'Average combat power'
	
def pokemonPerGeneration(data):
    """ Show the number of Pokemon per Generation in a pie chart"""
    plt.figure()
    data=data.drop_duplicates(subset='#')
    df=data.Generation.value_counts().sort_index() # sort by generation instead of counts
    df.plot(kind='pie', shadow=True, autopct=make_autopct(df))
    plt.title("Number of Pokémon per generation")
    			
def buildTeam(data):
	""" Try to implement an algorithm that finds the most efficient Team possible. A Team consists of six Pokemon. 
	An efficient Team covers a wide range of different Types while having as much combat power as possible. Try considering both the 
	primary and the secondary Type of a pokemon, if it has one.
	To make your team more realistically achievable, try using only non-legendary pokemon
	"""
	pokemon = data.sort_values(['Total'])
	pokemon = pokemon[~pokemon["Name"].str.contains("Mega", case = True)]
	pokemon = pokemon[~pokemon["Legendary"]]
	
	team = []
	teamTypes = []
	
	for p in reversed(pokemon.index):
		# check if primary type has not been picked yet
		# if a secondary type is present, check that type too
		if(pokemon.loc[p,"Type 1"] not in teamTypes and (pd.isnull(pokemon.loc[p,"Type 2"]) or pokemon.loc[p,"Type 2"] not in teamTypes)):
			# add member to the team
			team += [p]
			teamTypes += [pokemon.loc[p,"Type 1"]]
			if(not pd.isnull(pokemon.loc[p,"Type 2"])):
				teamTypes += [pokemon.loc[p,"Type 2"]]
				
			# stop when the teamsize of 6 is reached
			if(len(team) >= 6):
				return pokemon.ix[team]
				
def buildTeamBonus(data):
	""" This function should work in a similar way as buildTeam. Instead of trying to cover a wide array of types, you should build your team in a way 
	that your pokemon have effective types against a wide array of enemy types. Hint: Each type has one or more types that it is weak to (e.g. fire is weak against water).
	"""
	pokemon = data.sort_values(['Total'])
	pokemon = pokemon[~pokemon["Name"].str.contains("Mega", case = True)]
	pokemon = pokemon[~pokemon["Legendary"]]
	
	team = []
	uncoveredTypes = ["Bug","Electric","Fire","Grass","Normal","Rock","Dark","Fairy","Flying","Ground","Poison","Steel","Dragon","Fighting","Ghost","Ice","Psychic","Water"]
	
	# at first, only pokemon that are effective against a wide array of pokemon are accepted
	# this condition gets more loose the longer the program runs
	for acceptanceCriterion in [3,2,1,0]:
		for p in reversed(pokemon.index):
			# calculate all the types this pokemon is effective against
			effectiveAgainst = list(effectiveTypes[pokemon.loc[p,"Type 1"]])
			if(not pd.isnull(pokemon.loc[p,"Type 2"])):
				effectiveAgainst += effectiveTypes[pokemon.loc[p,"Type 2"]]
			
			# calculate number of uncovered types this pokemon is effective against
			effectCnt = len([t for t in effectiveAgainst if t in uncoveredTypes])

			# check if pokemon fulfills acceptance criterion, potentially add it to the team
			if (effectCnt>=acceptanceCriterion and p not in team):
				team += [p]
				uncoveredTypes = [t for t in uncoveredTypes if t not in effectiveAgainst]
				
			# stop when the teamsize of 6 is reached
			if(len(team) >= 6):
				return pokemon.ix[team]
			
			# if there are no uncovered types, acceptance criterion is set to 0 and the strongest available pokemon is picked
			if(len(uncoveredTypes) == 0):
				acceptanceCriterion = 0
				break
	

def main():
	filename = "pokemon.csv"   # file to read from
	data = pd.read_csv(filename)

	data = calculateTotalCombatPower(data)
	printCSV(data, "pokemon2.csv")
	topNPokemon(data, 20)
	topNNoMegaPokemon(data, 20)
	averagePowerPerGeneration(data)
	pokemonPerGeneration(data)
	print(buildTeamBonus(data))
	plt.show()

main()
